# distutils: language = c++
# cython: language_level = 3

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

from libc.stdint cimport int64_t, uint64_t
from libcpp.map cimport map as cpp_map
from libcpp.pair cimport pair
from libcpp.vector cimport vector

from usainboltz.cpp_simulator cimport CRule


cpdef int bounded_geometric(double, int, int);
cpdef int bounded_poisson(double, double, int, int);
cpdef int bounded_mset_distribution(list, list, int);


# CRule* is valid syntax wherever the cython parser expects a type, but not inside
# brackets e.g. when applying C++ templates.
# To work around this issue we define an alise for CRule*
ctypedef CRule* CRule_ptr

ctypedef struct wrule:
    CRule* rule
    vector[double] weights

cdef class Simulator:
    cdef CRule *c_epsilon
    cdef CRule *c_atom
    cdef cpp_map[int, CRule*] c_markers
    cdef cpp_map[int, CRule*] c_rules
    cdef object mapping

    cdef void grammar_to_c_struct(self, object, dict) except *
    cpdef list run(self, object, list)
    cdef wrule rule_to_c_struct(self, object, dict) except *
    cdef wrule union_to_c_struct(self, object, dict) except *
    cdef wrule product_to_c_struct(self, object, dict) except *
    cdef wrule seq_to_c_struct(self, object, int, int, dict) except *
    cdef wrule set_to_c_struct(self, object, int, int, dict) except *
    cdef wrule mset_to_c_struct(self, object, int, int, dict) except *

# Xoshiro stuff
cpdef void rng_seed(uint64_t)
cpdef double rng_double()
cpdef int64_t rng_i64(int64_t)
cdef void rng_get_state(uint64_t dest[4])
cdef void rng_set_state(const uint64_t s[4])



# coding: utf-8

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""

.. autosummary::

   binary_tree
   composition
   dyck_word

"""



# distutils: language = c++
# cython: language_level = 3

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

from libc.math cimport exp, pow, tgamma
from libcpp.utility cimport pair
from libcpp.vector cimport vector

from usainboltz.cpp_simulator cimport (
    CRule,
    bnd_geometric,
    bnd_mset_dist,
    bnd_poisson,
    c_simulate,
    get_state,
    initial_partition,
    make_atom,
    make_epsilon,
    make_marker,
    make_mset,
    make_product,
    make_ref,
    make_seq,
    make_set,
    make_union,
    next_partition,
    rand_double,
    rand_i64,
    seed,
    set_state,
)

from usainboltz.grammar import (
    Atom,
    Epsilon,
    Marker,
    MSet,
    Product,
    RuleName,
    Seq,
    Set,
    Symbol,
    Union,
)

# --- Wrapper on top of basic distributions implemented in C++ ------


cpdef int bounded_geometric(double d, int n, int m):
    return bnd_geometric(d,n,m)

cpdef int bounded_poisson(double a, double b, int c, int d):
    return bnd_poisson(a,b,c,d)

cpdef int bounded_mset_distribution(list weights, list arg_weights, int multiplicity):
    return bnd_mset_dist(weights, arg_weights, multiplicity)


# --- Wrapper on top of the C++ simulator ---------------------------


cdef double partition_weight(vector[pair[int, int]] part,
                             vector[double] arg_weight,
                             int j) except *:
    cdef double res = 1.0
    cdef int v, m, d = arg_weight.size()

    for v, m in part:
        if v * j < d:
            res *= pow(arg_weight[v * j], m)
            res /= tgamma(m + 1)
            res /= pow(v, m)

    return res


cdef class Simulator:
    def __init__(self, grammar, weights, mapping):
        self.mapping = mapping
        self.grammar_to_c_struct(grammar, weights)

    cdef void grammar_to_c_struct(self, object grammar, dict weights) except *:
        """Conversion of a Python grammar into a C++ grammar"""

        cdef pair[int, CRule*] binding

        # Allocate C rules for Epsilon() and Atom()
        self.c_epsilon = make_epsilon()
        self.c_atom = make_atom()

        # Allocate one C rule per Marker()
        for marker in grammar.markers():
            binding.first = self.mapping.marker_to_id[marker]
            binding.second = make_marker(binding.first)
            self.c_markers.insert(binding)

        # Allocate one C rule per RuleName()
        for rule_name in grammar.rules:
            binding.first = self.mapping.rule_name_to_id[rule_name]
            binding.second = make_ref(binding.first)
            self.c_rules.insert(binding)

        # Generate C rules for the right-hand-side of each grammar rule and bind them to
        # their rule name.
        cdef CRule *rule
        cdef wrule body
        for rule_name, rule_body in grammar.rules.items():
            rule = self.c_rules[self.mapping.rule_name_to_id[rule_name]]
            body = self.rule_to_c_struct(rule_body, weights)
            rule.set_ref_rule(body.rule)

    cpdef list run(self, object rulename, list max_sizes):
        cdef vector[int] ms = max_sizes
        cdef CRule* start = self.c_rules[self.mapping.rule_name_to_id[rulename]]
        return c_simulate(start, ms)

    cdef wrule rule_to_c_struct(self, object rule, dict weights) except *:
        cdef wrule res
        cdef int ls, us

        if isinstance(rule, Symbol):
            if isinstance(rule, Epsilon):
                res.rule = self.c_epsilon
            elif isinstance(rule, Atom):
                res.rule = self.c_atom
            elif isinstance(rule, Marker):
                res.rule = self.c_markers[self.mapping.marker_to_id[rule]]
            elif isinstance(rule, RuleName):
                res.rule = self.c_rules[self.mapping.rule_name_to_id[rule]]
            else:
                assert False
            res.weights = weights[rule]
            return res
        elif isinstance(rule, Union):
            return self.union_to_c_struct(rule.args, weights)
        elif isinstance(rule, Product):
            return self.product_to_c_struct(rule.args, weights)
        elif isinstance(rule, Seq):
            ls = 0 if rule.lower_size is None else rule.lower_size
            us = -1 if rule.upper_size is None else rule.upper_size
            return self.seq_to_c_struct(rule.arg, ls, us, weights)
        elif isinstance(rule, Set):
            ls = 0 if rule.lower_size is None else rule.lower_size
            us = -1 if rule.upper_size is None else rule.upper_size
            return self.set_to_c_struct(rule.arg, ls, us, weights)
        elif isinstance(rule, MSet):
            ls = 0 if rule.lower_size is None else rule.lower_size
            us = -1 if rule.upper_size is None else rule.upper_size
            return self.mset_to_c_struct(rule.arg, ls, us, weights)
        else:
            raise NotImplementedError(f"Operator {type(rule)} not yet supported.")

    cdef wrule union_to_c_struct(self, args, dict weights) except *:
        cdef wrule res, tmp
        cdef vector[vector[double]] probabilities
        cdef vector[CRule_ptr] c_args
        cdef int i, j, d = len(weights[Atom()]), nargs = len(args)

        c_args.clear()
        res.weights = [0.0] * d
        probabilities = [[0.0] * d for _ in range(nargs)]

        for i, arg in enumerate(args):
            tmp = self.rule_to_c_struct(arg, weights)
            c_args.push_back(tmp.rule)
            for j in range(1, d):
                probabilities[i][j] = tmp.weights[j]
                res.weights[j] += tmp.weights[j]
        for i in range(nargs):
            for j in range(1, d):
                if res.weights[j] > 0:
                    probabilities[i][j] /= res.weights[j]
        assert(c_args.size() > 1)
        res.rule = make_union(c_args, probabilities)
        return res

    cdef wrule product_to_c_struct(self, args, dict weights) except *:
        cdef wrule res, tmp
        cdef vector[CRule_ptr] c_args
        cdef int j, d = len(weights[Atom()])

        c_args.clear()
        res.weights = [1.0] * d

        for arg in args:
            tmp = self.rule_to_c_struct(arg, weights)
            c_args.push_back(tmp.rule)
            for j in range(1, d):
                res.weights[j] *= tmp.weights[j]
        res.rule = make_product(c_args)
        return res

    cdef wrule seq_to_c_struct(self, arg, int lower_size, int upper_size, dict weights) except *:
        cdef wrule res, c_arg
        cdef int j, d = len(weights[Atom()])
        cdef double ps_inv

        res.weights = [0.0] * d
        res.weights[0] = 1.0
        c_arg = self.rule_to_c_struct(arg, weights)

        for j in range(1, d):
            ps_inv = 1 / (1 - c_arg.weights[j])
            res.weights[j] = ps_inv
            res.weights[j] *= c_arg.weights[j] ** lower_size
            if upper_size > 0:
                res.weights[j] -= ps_inv * c_arg.weights[j] ** (upper_size + 1)
        res.rule = make_seq(
            c_arg.rule, res.weights, c_arg.weights, lower_size, upper_size
        )
        return res

    cdef wrule set_to_c_struct(self, arg, int lower_size, int upper_size, dict weights) except *:
        cdef wrule res, c_arg
        cdef int i, j, d = len(weights[Atom()])

        res.weights = [0.0] * d
        res.weights[0] = 1.0
        c_arg = self.rule_to_c_struct(arg, weights)

        for j in range(1, d):
            if upper_size > 0:
                for i in range(upper_size + 1):
                    res.weights[j] += c_arg.weights[j] ** i / tgamma(i + 1)
            else:
                res.weights[j] = exp(c_arg.weights[j])
                for i in range(lower_size):
                    res.weights[j] -= c_arg.weights[j] ** i / tgamma(i + 1)
        res.rule = make_set(
            c_arg.rule, res.weights, c_arg.weights, lower_size, upper_size
        )
        return res

    cdef wrule mset_to_c_struct(self, arg, int lower_size, int upper_size, dict weights) except *:
        cdef wrule res, c_arg
        cdef int i, j, d = len(weights[Atom()])
        cdef double acc
        cdef vector[vector[double]] partitions_weight
        cdef vector[pair[int, int]] partition

        res.weights = [0.0] * d
        res.weights[0] = 1.0
        c_arg = self.rule_to_c_struct(arg, weights)

        # M(z) = MSet_{lower_size <= . <= upper_size}(A)(z)
        if upper_size >= 0:
            if lower_size < upper_size:
                raise NotImplementedError("MSet with bounds other than `eq=...`")
            upper_size = min(d - 1, upper_size)
            lower_size = min(d - 1, lower_size)

            partition = initial_partition(lower_size)
            done = False
            while not done:
                part_weight = [0.0]
                for j in range(1, d):
                    # The weight of the current partition at z^j
                    acc = partition_weight(partition, c_arg.weights, j)
                    part_weight.append(acc)
                    res.weights[j] += acc
                partitions_weight.push_back(part_weight)
                done = next_partition(partition)
        elif lower_size > 0:
            raise NotImplementedError("Lower-bounded MSet with no upper bound")
        # M(z) = MSet(A)(z)
        else:
            for j in range(1, d):
                i = 1
                acc = 1.0
                while i * j < d:
                    acc *= exp(c_arg.weights[i * j] / i)
                    i += 1
                res.weights[j] = acc

        res.rule = make_mset(
            c_arg.rule,
            res.weights,
            c_arg.weights,
            partitions_weight,
            lower_size,
            upper_size
        )
        return res


# --- Wrapper on top of PRNG functions ------------------------------


cpdef void rng_seed(uint64_t s):
    seed(s)


cpdef double rng_double():
    return rand_double()


cpdef int64_t rng_i64(int64_t bound):
    return rand_i64(bound)


cdef void rng_get_state(uint64_t dest[4]):
    get_state(dest)


cdef void rng_set_state(const uint64_t s[4]):
    set_state(s)



# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""A Boltzmann sampler for alcohols

>>> from usainboltz import *
>>> from usainboltz.generator import rng_seed
>>> rng_seed(0xDEADBEEFB0175)  # For reproducibility

According to [FS2009] page 477, alcohols can been seen as rooted non-plane trees of
degree 3. The carbonic skeleton gives the tree structure and the carbon atom to which
the (unique) -OH group is attached is the root of the tree.
We thus use the :py:class:`~usainboltz.grammar.MSet` together with the :code:`eq=3`
constraint to specify these trees.

>>> carbon = Atom()
>>> A = RuleName("A")  # A for alcohol
>>> hydrogen = Marker("H")
>>> grammar = Grammar({A: hydrogen + carbon * MSet(A, eq=3)})
>>> grammar
{
  A : Union(H, Product(z, MSet(A, eq = 3)))
}

According to [FS2009], the singularity of the generating function of alcohols is near
0.35518174. Let us verify that the oracle finds a similar value.

>>> oracle = build_oracle(grammar)
>>> values = oracle.tuning(carbon, singular=True)
>>> print(f"{values[carbon][1]:.8f}")
0.35518174

The tree representation does not help a lot to visualise the molecule so let us define a
custom data structure for representing alcohols

>>> class Alcohol:
...     def __init__(self, *args):
...         main_builder = union_builder(self._init_hydrogen, self._init_carbon)
...         main_builder(*args)
...
...     def _init_hydrogen(self, epsilon):
...         self.element = "H"
...
...     def _init_carbon(self, args):
...         c, children = args
...         self.element = "C"
...         self.children = children
...
...     def to_dot(self):
...         # Print in graphviz format
...         print('graph {\n  node [shape="none", width=0.4, height=0.4]')
...         root_id = self._to_dot(0)
...         print('  root_node [label="OH"]')
...         print(f"  root_node -- node{root_id}")
...         print("}")
...
...     def _to_dot(self, counter):
...         # counter is the id of the last printed object
...         if self.element == "H":
...             counter += 1
...             print(f'  node{counter} [label="H"]')
...             return counter
...         else:
...             children_names = []
...             for c in self.children:
...                 counter = c._to_dot(counter)
...                 children_names.append(f"node{counter}")
...             counter += 1
...             print(f'  node{counter} [label="C"]')
...             print(f'  node{counter} -- {{{" ".join(children_names)}}}')
...             return counter

We defined the __init__ method of our class so that it can be passed directly to the
generator as a builder. Let us to so and run the random sampler:

>>> generator = Generator(grammar)
>>> generator.set_builder(A, Alcohol)  # pass the class constructor directly
>>> res = generator.sample((10, 15))

We generated an alcohol with the following elements:

>>> SUB = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
>>> nb_carbons = str(res.sizes[carbon]).translate(SUB)
>>> nb_hydrogens = str(res.sizes[hydrogen] + 1).translate(SUB)
>>> print(f"C{nb_carbons}OH{nb_hydrogens}")
C₁₅OH₃₂

Using the printer we defined above, we can print a DOT reprensetation of the generated
alcohol:

>>> res.obj.to_dot()  # doctest: +ELLIPSIS
graph {
  node [shape="none", width=0.4, height=0.4]
  node1 [label="H"]
  node2 [label="H"]
  node3 [label="H"]
  node4 [label="C"]
  node4 -- {node1 node2 node3}
  node5 [label="H"]
...

Using the neato rendered of the graphviz library, we get the following picture:

.. image:: ../_static/img/alcohol.svg
   :height: 480px
   :width: 480px
   :align: center
"""

from usainboltz import Atom, Epsilon, Generator, Grammar, MSet, RuleName

if __name__ == "__main__":
    carbon = Atom()
    A = RuleName("A")  # A for alcohol
    hydrogen = Epsilon()
    grammar = Grammar({A: hydrogen + carbon * MSet(A, eq=3)})

    generator = Generator(grammar)
    for _ in range(20):
        res = generator.sample((10, 20))
        print(res.sizes[carbon])



# coding: utf-8

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

"""
Various oracle implementations for Boltzmann sampling.

Oracles are used to get (often approximate) values of generating functions.
Thanks to the symbolic method, functional equations can be derived from
grammar specifications. This module implements some mechanics to approximate
generating functions based on these equations.

Currently three oracles are implemented:

- :class:`OracleFromDict` does nothing. It takes a dictionary of already computed
  values from the user and wraps them as an :class:`Oracle`.
  This dictionary must have all the :class:`~usainboltz.grammar.Symbol`s of the grammar
  as keys and must be of one of the two following form:
  - Either it maps the symbols of the grammar to floating point values, in which
    case these values are interpreted as the values of the corresponding generating
    functions at a common point `z`.
  - Or it maps them to an array of floating point values, in which case the `j`-th
    entry of each array is interpreted as the value of the corresponding generating
    function at `z^j`.
- :class:`OracleFromPaganini` use the oracle Pagagnini of [BBD2018]_
- :class:`OracleFromNewtonGF` use the oracle NewtonGF of [PSS2012]_
  (only when Sagemath and Maple installations are present)

The entry point of these algorithms is the function oracle which determines
the oracle to use given its inputs.

AUTHORS:
- Matthieu Dien (2019): initial version
- Martin Pépin (2019): initial version

.. rubric:: References

.. [BBD2018] Maciej Bendkowski, Olivier Bodini, Sergey Dovgal, 2018,
   Polynomial tuning of multiparametric combinatorial samplers, ANALCO
.. [PSS2012] Carine Pivoteau, Bruno Salvy, Michèle Soria, 2012
   Algorithms for Combinatorial Structures: Well-founded systems and
   Newton iterations, Journal of Combinatorial Theory
"""

import sys
from functools import partial, reduce
from typing import Callable, Dict, List, NamedTuple, Optional, Tuple, Union as TypeUnion

import paganini as pg
from cvxpy import SolverError

from usainboltz.grammar import (
    Atom,
    Cycle,
    Epsilon,
    Grammar,
    Marker,
    MSet,
    Product,
    Rule,
    RuleName,
    Seq,
    Set,
    Symbol,
    UCycle,
    Union,
)

__all__ = ["build_oracle", "OracleFromPaganini", "OracleFromDict"]


class Oracle:
    """The super class of all oracles.

    Should not be instantiated directly.

    The ``tuning`` method must compute a suitable z and the corresponding values of the
    generating functions of the system.
    Sub classes may implement any interface for the tuning method as long as their
    output ``d`` satisfy the following requirements:

    - ``d[Atom()]`` is the list ``[1, z, z**2, z**3, ..., z**k]`` for some
      floating point value ``z`` et some positive integer ``k``;
    - for each symbol ``S`` of the grammar, ``d[S]`` is the list
      ``[1, S(z), S(z**2), S(z**3), ..., S(z**k)]`` where ``S(z**j)`` denotes the
      generating function for ``S`` (ordinary or exponential depending on the
      context) evaluated at ``z**j``;
    - all the lists appearing in this dictionary should have the same length
      (``k+1``).

    Most of the time ``k=1`` is enough but when Pólya operators like
    :py:class:`~usainboltz.grammar.MSet` appear in the grammar, more values are
    necessary.
    In such situations, any generation function ``S(z**j)`` with ``j > k`` will be
    approximated by zero.
    """

    tuning: Callable[..., Dict[Symbol, List[float]]]


def build_oracle(sys, **kargs) -> Oracle:
    """Build different oracles given different inputs

    EXAMPLES::
        >>> from usainboltz import *

        >>> leaf, z = Marker("leaf"), Atom()
        >>> B = RuleName("B")
        >>> g = Grammar({B: leaf + z * B * B})
        >>> build_oracle(g)
        OracleFromPaganini({
          B : Union(leaf, Product(z, B, B))
        })

        >>> build_oracle({'B': 2, 'z': 1./4})
        OracleFromDict({'B': [1.0, 2], 'z': [1.0, 0.25]})
    """

    if isinstance(sys, Grammar):
        return OracleFromPaganini(sys)
    elif isinstance(sys, dict):
        return OracleFromDict(sys)

    raise TypeError(
        "I do not know what to do with a system of type {}".format(type(sys))
    )


# ------------------------------------------------------- #
# Interface with Paganini
# ------------------------------------------------------- #


class _PaganiniEnv(NamedTuple):
    """Contextual information for interacting with paganini

    This stores the necessary information to extract the tuning result from paganini
    after having converted our grammar to their formalism.

    Args:
        variables: A mapping from (our) symbols to their corresponding paganini
            expression

        diagonals: A dictionary mapping paganini variables to their corresponding
            diagonals in paganini's internal representation. For instance, for the
            specification `T = z * MSet(T)`, paganini creates new variables v2, v3, v4,
            … representing T(z^2), T(z^3), T(z^4), … which it calls "diagonals". The
            dictionary below stores, for each symbol T and for each integer j, the
            paganini variable representing the j-th diagonal T(z^j) of T. This
            dictionary is empty when no Pólya operator is used in the specification.

        series_truncate: When Pólya operators are used, diagonals of degree larger or
            equal to `series_truncate` are approximated to zero.

    """

    variables: Dict[Symbol, pg.Expr]
    diagonals: Dict[pg.Variable, Dict[int, pg.Variable]]
    series_truncate: int

    def to_values(self) -> Dict[Symbol, List[float]]:
        """Convert paganini's output in the format expected by the generator.

        Examples:
            >>> from usainboltz import *
            >>> z, u, eps = Atom(), Marker("u"), Epsilon()
            >>> A, B =  RuleName("A"), RuleName("B")

            Some random algebraic specification.
            There is no Pólya operator so the ouput contains only arrays of size 2.

            >>> grammar = Grammar({A: z + B * A**2, B: z * u + eps})
            >>> values = OracleFromPaganini(grammar).tuning(A, singular=True)
            >>> print(len(values[z]))
            2
            >>> print(f"{values[z][1]:.8f}")
            0.20710678

            The specification of non-plane (Pólya) n-ary trees:

            >>> grammar = Grammar({A: z * MSet(A)})
            >>> values = OracleFromPaganini(grammar).tuning(A, singular=True)
            >>> print((len(values[z]), len(values[A])))
            (30, 30)
            >>> for j in range(1, 10):
            ...     print(f"{values[A][j]:.8f}")
                1.00000000
                0.13148788
                0.04035050
                0.01327774
                0.00445234
                0.00150187
                0.00050761
                0.00017168
                0.00005808
            >>> abs(values[z][1] ** 7 - values[z][7]) < 1e-9
            True
        """
        values: Dict[Symbol, List[float]] = dict()
        diags = self.diagonals if self.diagonals else None
        series_truncate = self.series_truncate if diags is not None else 2

        values[Epsilon()] = [1.0] * series_truncate

        for symbol, pg_expr in self.variables.items():
            # XXX. The atom seems to have a special treatment in paganini
            if symbol == Atom():
                values[symbol] = [pg_expr.value**j for j in range(series_truncate)]

            elif isinstance(pg_expr, pg.Variable):
                if diags is not None:
                    values[symbol] = [1.0]
                    diag = diags[pg_expr]
                    for j in range(1, series_truncate):
                        values[symbol].append(diag[j].value if j in diag else 0.0)
                else:
                    values[symbol] = [1.0, pg_expr.value]

            else:
                # Then pg_expr should be a constant.
                assert pg_expr.is_constant
                values[symbol] = [pg_expr.coeff] * series_truncate

        return values


def _to_paganini_constraint(
    lower_size: Optional[int], upper_size: Optional[int]
) -> Optional[pg.Constraint]:
    r"""Translates constraints over :py:class:`usainboltz.grammar.IteratedRule`
    into :py:class:`paganini.Constraint`"""
    if lower_size is not None and upper_size is not None:
        if lower_size == upper_size:
            return pg.eq(lower_size)
        else:
            raise ValueError("Paganini does not support mutliple constraints.")
    elif lower_size is not None:
        return pg.geq(lower_size)
    elif upper_size is not None:
        return pg.leq(upper_size)
    else:
        return None


def _to_paganini_rule(env: _PaganiniEnv, rule: Rule) -> pg.Expr:
    r"""Recursively translates :py:class:`usainboltz.grammar.Rule`
    into :py:class:`paganini.Expr` using the environnement `env` to
    set the expectations of pagagnini `Variables`
    """
    if isinstance(rule, Epsilon):
        return pg.Expr()
    elif isinstance(rule, Symbol):  # means Atom, Marker or RuleName
        return env.variables[rule]
    elif isinstance(rule, Union):
        return reduce(
            lambda x, y: x + y, map(partial(_to_paganini_rule, env), rule.args)
        )
    elif isinstance(rule, Product):
        return reduce(
            lambda x, y: x * y, map(partial(_to_paganini_rule, env), rule.args)
        )
    elif isinstance(rule, Seq):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.Seq(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, Set):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.Set(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, Cycle):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.Cyc(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, MSet):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.MSet(_to_paganini_rule(env, rule.arg), constraint)
    elif isinstance(rule, UCycle):
        constraint = _to_paganini_constraint(rule.lower_size, rule.upper_size)
        return pg.UCyc(_to_paganini_rule(env, rule.arg), constraint)
    else:
        raise TypeError("second argument should be of type usainboltz.grammar.Rule")


def _to_paganini_spec(
    grammar: Grammar, weights: Dict[Symbol, float]
) -> Tuple[_PaganiniEnv, pg.Specification]:
    r"""Convert a :py:class:`usainboltz.grammar.Grammar` with `weights`
    over symbols into a map :py:class:`usainboltz.grammar.Rule` ->
    :py:class:`paganini.Expr` and a :py:class:`paganini.Specification`

    Example:
        >>> from usainboltz import *
        >>> eps = Epsilon();  z = Atom(); B = RuleName()
        >>> grammar = Grammar({B: eps + z * B * B})

        >>> from usainboltz.oracle import _to_paganini_spec
        >>> env, spec = _to_paganini_spec(grammar, {z:1000})
        >>> env.variables[z], env.variables[B]
        (var1, var0)
        >>> spec
        var0 = 1  + var1^1 var0^2
    """

    spec = pg.Specification(series_truncate=30)
    env = _PaganiniEnv(
        variables=dict(),
        diagonals=spec._diag,  # XXX. Undocumented paganini feature.
        series_truncate=spec._series_truncate,  # XXX. Undocumented paganini feature.
    )

    for rulename in grammar.rules.keys():
        env.variables[rulename] = pg.Variable(weights.get(rulename))

    env.variables[Atom()] = pg.Variable(weights.get(Atom()))

    for m in grammar.markers():
        if m in weights:
            env.variables[m] = pg.Variable(weights[m])
        else:
            env.variables[m] = pg.Expr()

    for rulename, rule in grammar.rules.items():
        spec.add(env.variables[rulename], _to_paganini_rule(env, rule))

    return env, spec


class OracleFromPaganini(Oracle):
    r"""Build an oracle using
    `paganini <https://github.com/maciej-bendkowski/paganini>`_ : a package
    developped by M. Bendkowski and S. Dovgal

    .. todo:: add the reference to the paper
    """

    def __init__(self, grammar: Grammar):
        r"""Automatically build an oracle from a grammar using paganini.

        Args:
            grammar: the grammar for which to build the oracle.
        """
        super(OracleFromPaganini, self).__init__()
        self.grammar = grammar
        self.pg_method = pg.Method.FORCE

    def tuning(
        self, rule: RuleName, expectations=None, singular=False
    ) -> Dict[Symbol, List[float]]:
        r"""Run the oracle's algorithm

        Args:
          rule: the targeted symbol to tune w.r.t. ``expectations``

          expectations: a mapping between grammar markers and their targeted
              expectations after tuning

          singular: if ``True`` run the singular tuner (infinite expected size) else run
              a tuner targeting the given ``expectations``

        Returns:
            values: a mapping between grammar symbols and their weights in the Boltzmann
                model

        Note:
            If ``singular`` is ``True``, then ``expectations`` should
            be should be a dictionnary with grammar names (but the
            atom) as keys and ratios (between 0 and 1) as values.

           Else, ``expectations`` should be a dictionnary with keys as
            grammar names (but ``rule``) and the values should be
            integers.

        Examples:
           Use of the singular tuner for a binary tree grammar (size
           is the number of internal nodes):

           >>> from usainboltz import *
           >>> eps = Epsilon();  z = Atom(); B = RuleName()
           >>> grammar = Grammar({B: eps + z * B * B})
           >>> o = OracleFromPaganini(grammar)
           >>> values = o.tuning(B, singular=True)
           >>> abs(values[z][1]-0.25) < 1e-12
           True
           >>> abs(values[B][1]-2.) < 1e-12
           True

           As we expect ``z`` should be :math:`\frac14` (the
           singularity of :math:`\frac{1-\sqrt{1-4z}}{2}`)
        """

        if expectations is None:
            expectations = dict()

        if rule in expectations:
            raise ValueError("You must not set any constraint over {}".format(rule))

        # Check the values of expectations w.r.t. singular or not
        if singular:
            if any(x < 0 or x > 1 for x in expectations.values()):
                raise ValueError("Ratios must be between 0 and 1")
        elif all(x < 1 for x in expectations.values()):
            raise ValueError("The expected numbers must be greater than 1")

        env, spec = _to_paganini_spec(self.grammar, expectations)

        env.variables[rule].set_expectation(None)
        for k, v in expectations.items():
            env.variables[k].set_expectation(v)

        try:
            if singular:
                spec.run_singular_tuner(env.variables[Atom()], method=self.pg_method)
            else:
                spec.run_tuner(env.variables[rule], method=self.pg_method)
        except SolverError:
            raise ValueError("Your expected sizes can not be targeted")

        return env.to_values()

    def __repr__(self):
        return "OracleFromPaganini({})".format(self.grammar)


# ------------------------------------------------------- #
# Convertion of a python dictionnary into an oracle
# ------------------------------------------------------- #


class OracleFromDict(Oracle):
    dict_: Dict[Symbol, List[float]]

    def __init__(self, d: TypeUnion[Dict[Symbol, List[float]], Dict[Symbol, float]]):
        super(OracleFromDict, self).__init__()
        if not (isinstance(d, dict)):
            raise TypeError("The argument should be a dictionnary")

        # Take one arbitrary value of the dictionary
        for v in d.values():
            if isinstance(v, (float, int)):
                self.dict_ = {
                    key: [1.0, value] for key, value in d.items()  # type: ignore
                }
            else:
                # mypy is not clever enough to see that this is well-typed (is it?).
                self.dict_ = d  # type: ignore
            return
        raise ValueError("OracleFromDict received an empty dictionary")

    def tuning(self, *args, **kwargs) -> Dict[Symbol, List[float]]:
        return self.dict_

    def __repr__(self):
        return "OracleFromDict({})".format(self.dict_)


# ------------------------------------------------------- #
# Interface with combstruct (maple)
# ------------------------------------------------------- #


def _to_combstruct_constraint(
    lower_size: Optional[int], upper_size: Optional[int]
) -> str:
    r"""Translates sizes constraint over
    :py:class:`usainboltz.grammar.IteratedRule` into `combstruct`
    constraints.

    Examples:
        >>> from usainboltz import * # doctest: +SKIP
        >>> z = Atom(); S = RuleName("S") # doctest: +SKIP
        >>> o = OracleFromNewtonGF(Grammar({S: Seq(z, leq=3)})) # doctest: +SKIP
        >>> o.combstruct_spec # doctest: +SKIP
        '{S = Seq(Atom, card <= 3)}'
        >>> o = OracleFromNewtonGF(Grammar({S: Seq(z, geq=3)})) # doctest: +SKIP
        >>> o.combstruct_spec # doctest: +SKIP
        '{S = Seq(Atom, 3 <= card)}'
        >>> o = OracleFromNewtonGF(Grammar({S: Seq(z, eq=3)})) # doctest: +SKIP
        >>> o.combstruct_spec # doctest: +SKIP
        '{S = Seq(Atom, card = 3)}'
    """

    x = (lower_size, upper_size)
    if x == (None, None):
        return ""
    elif x == (lower_size, None):
        return f", {lower_size} <= card"
    elif x == (None, upper_size):
        return f", card <= {upper_size}"
    else:
        if lower_size == upper_size:
            return f", card = {upper_size}"
        else:
            print(
                "WARNING: NewtonGF may not handle constraints "
                + "over lower and upper sizes",
                file=sys.stderr,
            )
            return f", card <= {upper_size}, {lower_size} <= card"


def _to_combstruct_rule(rule: Rule) -> str:
    r"""Recursively translates :py:class:`~usainboltz.grammar.Rule`
    into `combstruct[specification]`
    """

    if isinstance(rule, Epsilon):
        return "Epsilon"
    elif isinstance(rule, Atom):
        return "Atom"
    elif isinstance(rule, Marker):
        return "Epsilon"
    elif isinstance(rule, RuleName):
        return str(rule)
    elif isinstance(rule, Union):
        return "Union(" + ",".join(map(_to_combstruct_rule, rule.args)) + ")"
    elif isinstance(rule, Product):
        return "Prod(" + ",".join(map(_to_combstruct_rule, rule.args)) + ")"
    elif isinstance(rule, Seq):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Sequence({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, Set):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Set({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, Cycle):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Cycle({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, MSet):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Set({_to_combstruct_rule(rule.arg)}{constraint})"
    elif isinstance(rule, UCycle):
        constraint = _to_combstruct_constraint(rule.lower_size, rule.upper_size)
        return f"Cycle({_to_combstruct_rule(rule.arg)}{constraint})"
    #  elif isinstance(rule, PSet):
    #  raise NotImplementedError
    else:
        raise TypeError("first argument should be of type usainboltz.grammar.Rule")


def _to_combstruct_spec(grammar: Grammar) -> str:
    spec = set()

    for marker in grammar.markers():
        spec.add(f"{marker}=Epsilon")

    for rulename, rule in grammar.rules.items():
        spec.add(f"{rulename} = {_to_combstruct_rule(rule)}")

    return "{" + ",".join(spec) + "}"


newtongf_installed: bool = False

try:
    from sage.interfaces.maple import maple

    maple.load("combstruct")
    maple.load("NewtonGF")
    newtongf_installed = True
    __all__.append("OracleFromNewtonGF")
except ImportError:
    pass
except RuntimeError:
    pass


class OracleFromNewtonGF(Oracle):
    r"""Build an orcale using `NewtonGF
    <http://perso.ens-lyon.fr/bruno.salvy/software/the-newtongf-package/>`_

    Note:
        If you want to use OracleFormNewtonGF, please install the
        NewtonGF library for Maple into you Maple lib/ directory,
        or precise a lib/ directory to Maple i.e.
        maple('libname:=\"NewtonGF_path\",libname:')
    """

    def __init__(self, grammar: Grammar):
        r"""Build an oracle from a grammar"""

        if not newtongf_installed:
            raise ImportError(
                "You cannot use OracleFromNewtonGF because your "
                + "Sagemath, Maple or NewtonGF installation is not "
                + "consistent."
            )

        self.grammar = grammar
        self.combstruct_spec = _to_combstruct_spec(grammar)

    def tuning(
        self, rule: Symbol, expectations=None, singular=False
    ) -> Dict[Symbol, List[float]]:
        r"""Run the oracle's algorithm

        Args:
            rule: the targeted symbol to tune w.r.t. ``expectations``

            expectations: a mapping between grammar symbols and their targeted
                expectations after tuning

          singular: if ``True`` run the singular tuner (infinite expected size) else run
            a tuner targeting the given ``expectations``

        Returns:
            values: a mapping between grammar symbols and their weights in the Boltzmann
                model

        Note:
            In the case of NewtonGF, only **univariate**
            specifications are supported, so:
            * if `expectations` is provided it must have a key corresponding
            to :py:class:`~usainboltz.grammar.Atom()`,
            * the markers (see :py:class:`~usainboltz.grammar.Marker`)
            will not be handled.

        Examples:
           Use of the singular tuner for a binary tree grammar (size
           is the number of internal nodes):

           >>> from usainboltz import *
           >>> eps = Epsilon();  z = Atom(); B = RuleName()
           >>> grammar = Grammar({B: eps + z * B * B})
           >>> o = OracleFromNewtonGF(grammar) # doctest: +SKIP
           >>> values = o.tuning(B,singular=True) # doctest: +SKIP
           >>> values[z] # doctest: +SKIP
           0.25
           >>> values[B] # doctest: +SKIP
           1.999999988

           As we expect ``z`` should be :math:`\frac14` (the
           singularity of :math:`\frac{1-\sqrt{1-4z}}{2}`)

        """

        # To remove if we do the choice to autorize only one atom per grammar
        if len(self.grammar.markers()) > 0:
            print(
                "NewtonGF deals only with univariate specifications "
                + "and there are several markers in your one",
                file=sys.stderr,
            )

        labeled = "labeled" if self.grammar.labelled else "unlabeled"

        if not singular and (expectations is None or Atom() not in expectations):
            raise ValueError(
                "You must set an expectations over the size of the structure"
            )

        rho = 0
        if singular:
            rho = maple(f"Radius({self.combstruct_spec}, {labeled})")
        else:
            rho = maple(
                f"BoltzmannParameter({self.combstruct_spec}, {labeled}, {rule}, "
                + f"{expectations[Atom()]})"
            )

        oracle = maple(
            f"NumericalNewtonIteration({self.combstruct_spec}, {labeled})({rho})"
        )
        table = str.maketrans({"=": ":", "[": "{", "]": "}"})

        self.grammar_symbols = {
            str(symbol): symbol
            for symbol in self.grammar.markers() | set(self.grammar.rules)
        }

        values = eval(str(oracle).translate(table), globals(), self.grammar_symbols)
        values[Atom()] = float(rho)

        values_list = {symbol: [1.0, value] for symbol, value in values.items()}

        return values_list

    def __repr__(self):
        return "OracleFromNewtonGF({})".format(self.grammar)



# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

"""RNA secondary structures"""

from typing import List, TextIO, Tuple, Union as TypeUnion

from usainboltz import (
    Atom,
    Epsilon,
    Generator,
    Grammar,
    Marker,
    RuleName,
    union_builder,
)
from usainboltz.generator import rng_seed

B, S, z = RuleName("B"), RuleName("S"), Atom()
A, U, C, G = Marker("A"), Marker("U"), Marker("C"), Marker("G")


grammar = Grammar(
    {S: B * z * (S + Epsilon()) + B * z * S * z * (S + Epsilon()), B: A + U + C + G}
)
generator = Generator(grammar, S)


dual = {A: U, U: A, C: G, G: C}


def identity(x):
    return x


# ---
# RNA secondary structures as sequences of bases with offsets
# ---


# The Python type of sequences of bases.
SEQ = List[TypeUnion[str, Tuple[str, int]]]

# Tagged union of a SEQ or an Epsilon
SEQ_E = Tuple[int, TypeUnion[SEQ, str]]


def build_seq_empty(_: str) -> SEQ:
    return []


build_seq_s_or_e = union_builder(identity, build_seq_empty)


def build_seq_prefix(t: Tuple[str, SEQ_E]) -> SEQ:
    base, z, se = t
    return [base] + build_seq_s_or_e(se)


def build_seq_matching(t: Tuple[str, SEQ, str, SEQ_E]) -> SEQ:
    base, z, s, z, se = t
    offset = len(s) + 1
    return [(base, offset)] + s + [dual[base]] + build_seq_s_or_e(se)


build_seq_S = union_builder(build_seq_prefix, build_seq_matching)


# ---
# RNA secondary structures as forests
# ---


class Counter:
    def __init__(self):
        self.n = 0

    def next(self) -> int:
        r = self.n
        self.n += 1
        return r


# the type of forests
FOREST = List["Tree"]


def forest_str(f: FOREST) -> str:
    return "[{}]".format(", ".join(map(str, f)))


def forest_dot(f: FOREST, counter: Counter, file: TextIO) -> List[int]:
    return [tree.dot(counter, file) for tree in f]


class Tree:
    def __init__(self, children: FOREST):
        self.children = children

    def dot(self, counter, file) -> int:
        my_id = counter.next()
        children_ids = forest_dot(self.children, counter, file)
        file.write(f'  {my_id} ["shape"="point"]\n')
        for child_id in children_ids:
            file.write(f"  {my_id} -> {child_id}\n")
        return my_id

    def __str__(self):
        return f"Tree({forest_str(self.children)})"


def build_forest_prefix(t: Tuple[str, str, FOREST]) -> FOREST:
    _, _, s = t
    return s


def build_forest_matching(t: Tuple[str, str, FOREST, str, FOREST]) -> FOREST:
    _, _, s, _, sp = t
    return [Tree(s)] + sp


def build_forest_empty(_) -> FOREST:
    return []


build_forest_S = union_builder(build_forest_prefix, build_forest_matching)
build_forest_Sp = union_builder(identity, build_forest_empty)


# ---
# The order to RNA secondary structures
# ---


def order_prefix(t):
    _, _, s = t
    return s


def order_matching(t):
    _, _, s, _, sp = t
    order_s, _ = s
    order_sp, flag_sp = sp
    if order_s < order_sp:
        return sp
    if order_s > order_sp:
        return (order_s, True)
    if flag_sp:
        return (order_sp + 1, False)
    else:
        return (order_sp, True)


def order_empty(_):
    return (1, False)


order_S = union_builder(order_prefix, order_matching)
order_Sp = union_builder(identity, order_empty)


if __name__ == "__main__":
    SEED = 0x8FFF2FAD00DC576B
    SIZE = 15

    print(f"=> Using seed: 0x{SEED:x}")

    print("=> Grammar:")
    print(grammar)

    print("=> Using the default builders:")
    rng_seed(SEED)
    res = generator.sample((SIZE, SIZE))
    print(res.obj)

    print("=> A large RNA secondary structure:")
    rng_seed(SEED)
    res = generator.sample((10000, 15000))
    print(res.sizes)

    print("=> Using expectations:")
    rng_seed(SEED)
    generator2 = Generator(grammar, expectations={A: 3000, z: 10000})
    res = generator.sample({z: (10000, 15000), A: (2800, 15000)})
    print(res.sizes)

    print("=> Using the sequence builders:")
    rng_seed(SEED)
    generator.set_builder(S, build_seq_S)
    res = generator.sample((SIZE, SIZE))
    print(res.obj)

    exit(0)

    print("=> Using the forest builders:")
    generator.set_builder(S, build_forest_S)
    rng_seed(SEED)
    res = generator.sample((SIZE, SIZE))
    print(forest_str(res.obj))
    with open("forest.dot", "w") as file:
        file.write("digraph G {\n")
        forest_dot(res.obj, Counter(), file)
        file.write("}\n")

    print("=> Using the order builders:")
    generator.set_builder(S, order_S)
    rng_seed(SEED)
    res = generator.sample((SIZE, SIZE))
    print(res.obj)



# coding: utf-8

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

from usainboltz import Atom, Generator, Grammar, Marker, RuleName, union_builder

# ---
# S-expressions to represent any kind of expression
# ---


class Sexpr:
    def __init__(self, constructor, *args):
        self.constructor = constructor
        self.args = args

    def __eq__(self, other):
        return (
            isinstance(other, Sexpr)
            and other.constructor == self.constructor
            and other.args == self.args
        )

    def __str__(self):
        if self.args:
            args = ", ".join(map(str, self.args))
            return "({}, {})".format(self.constructor, args)
        else:
            return str(self.constructor)

    def size(self):
        return 1 + sum(arg.size() for arg in self.args)


# ---
# Regular expressions
# ---

# UsainBoltz grammar
z = Atom()
a, b, eps = Marker("a"), Marker("b"), Marker("eps")
L = RuleName()
grammar = Grammar(rules={L: z * (a + b + eps + L + L * L + L * L)})
generator = Generator(grammar, L, singular=True)

# Expression constructors
atom_a = Sexpr("a")
atom_b = Sexpr("b")
atom_eps = Sexpr("eps")


def star(expr):
    return Sexpr("*", expr)


def concat(x, y):
    return Sexpr(".", x, y)


def union(x, y):
    return Sexpr("+", x, y)


pattern = Sexpr("*", atom_a, atom_b)


def is_pattern(expr):
    a = Sexpr("a")
    b = Sexpr("b")
    return expr == Sexpr("*", a, b) or expr == Sexpr("*", b, a)


def build_atom(name):
    return Sexpr(name)


def build_star(expr):
    # Simplification: (expr)** = (expr)*
    if expr.constructor == "*":
        return expr
    # Simplification: epsilon* = epsilon
    elif expr == atom_eps:
        return expr
    # Simplification (expr + epsilon)* = (expr)*
    elif expr.constructor == "+":
        epsilon = Sexpr("eps")
        lhs, rhs = expr.args
        if lhs == epsilon:
            return star(rhs)
        elif rhs == epsilon:
            return star(lhs)
        else:
            return star(expr)
    else:
        return star(expr)


def build_union(args):
    lhs, rhs = args
    # Simplification: (a+b)* + expr = (a+b)*
    if is_pattern(lhs) or is_pattern(rhs):
        return pattern
    else:
        return union(lhs, rhs)


def build_concat(args):
    return concat(*args)


def build(tupl):
    build_arg = union_builder(
        build_atom, build_atom, build_atom, build_star, build_union, build_concat
    )
    z, arg = tupl
    return build_arg(arg)


generator.set_builder(L, build)



# coding: utf-8

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""
Context-free grammars for Boltzmann generation.
===============================================

Grammars use the basic operators of the symbolic method of analytic
combinatorics to specify labelled or unlabelled combinatorial classes. For
instance, binary trees can be specified by ``B = leaf + Z * B * B`` which,
using the syntax implemented in this module, looks like:

Examples:
    >>> z, leaf = Atom(), Marker("leaf")
    >>> B = RuleName("B")

    >>> # Using the explicit syntax
    >>> Grammar({B: Union(leaf, Product(z, B, B))})
    {
      B : Union(leaf, Product(z, B, B))
    }
    >>> # Using the syntactic sugar + and * for unions and products
    >>> Grammar({B: leaf + z * B * B})
    {
      B : Union(leaf, Product(z, B, B))
    }


Note that we make the difference between

- terminal symbols of the grammar: :py:class:`Atom` (``leaf`` and ``z`` in the
  above example).
- non-terminal symbols: :py:class:`RuleName` (``B`` in the above example).

The most basic constructions of the symbolic method are the Cartesian product
:py:class:`Product` and the disjoint union :py:class:`Union`. In the example of
binary trees, disjoint union is used to decompose the class between leaves and
internal nodes. The Cartesian product is used to define internal nodes as a
pair of children, the ``z`` appearing at each internal node means that each
node counts as :math:`1` in the size of a tree.

The exhaustive list of supported constructions is given below. Detailed
explanations about each of these operators can be found on the `Wikipedia page
<https://en.wikipedia.org/wiki/Symbolic_method_(combinatorics)>`_ of the
symbolic method and in [FS2009]_.

Here are two other examples of specifications, for general plane trees,
illustrating the use of the :class:`Seq` (sequence) operator and the use of
multi-rules grammars:

Examples:
    >>> z = Atom()
    >>> T, S = RuleName("T"), RuleName("S")

    >>> # With Seq
    >>> Grammar({T: z * Seq(T)})
    {
      T : Product(z, Seq(T))
    }

    >>> # Equivalent definition without Seq
    >>> nil = Epsilon()
    >>> Grammar({
    ...     T: z * S,
    ...     S: nil + T * S,
    ... })
    {
      S : Union(epsilon, Product(T, S)),
      T : Product(z, S)
    }

.. todo:: Write about labeling

AUTHORS:

- Matthieu Dien (2019): initial version

- Martin Pépin (2019): initial version

.. rubric:: References

.. [FS2009] Philippe Flajolet, Robert Sedgewick, 2009, Analytic combinatorics,
   Cambridge University Press.
"""

from functools import reduce
from typing import Dict, List, Optional, Set as PythonSet, Union as TypeUnion

__all__ = (
    "Marker",
    "Epsilon",
    "Atom",
    "Cycle",
    "Grammar",
    "MSet",
    "Product",
    "RuleName",
    "Seq",
    "Set",
    "Union",
    "UCycle",
)


# ------------------------------------------------------- #
# Fresh names generation
# ------------------------------------------------------- #


class _NameGenerator:
    def __init__(self, prefix: str):
        self.prefix = prefix
        self.counter = 0

    def fresh(self) -> str:
        fresh_name = "{}_{}".format(self.prefix, self.counter)
        self.counter += 1
        return fresh_name


_symbol_name_generator = _NameGenerator("_symbol")


# ------------------------------------------------------- #
# Grammar implementation
# ------------------------------------------------------- #


class Rule:
    """The super class of all grammar rules.

    Should not be instantiated directly.
    """

    def markers(self) -> PythonSet["Marker"]:
        raise NotImplementedError("Sub classes of Rule should implement self.markers()")

    def __add__(self, rhs: "Rule") -> "Rule":
        """Disjoint union of rules using the + syntax.

        Examples:
            >>> a, b, c, d = Marker("a"), Marker("b"), Marker("c"), Marker("d")

            >>> a + b
            Union(a, b)

            >>> Union(a, b) + c
            Union(a, b, c)

            >>> a + Union(b, c)
            Union(a, b, c)

            >>> Union(a, b) + Union(c, d)
            Union(a, b, c, d)
        """
        # ensure not to build unions of unions
        lhs_terms = self.args if isinstance(self, Union) else [self]
        rhs_terms = rhs.args if isinstance(rhs, Union) else [rhs]
        return Union(*(lhs_terms + rhs_terms))

    def __mul__(self, rhs: "Rule") -> "Rule":
        """Cartesian product of rules using the * syntax.

        Examples:
            >>> a, b, c, d = Marker("a"), Marker("b"), Marker("c"), Marker("d")

            >>> a * b
            Product(a, b)

            >>> Product(a, b) * c
            Product(a, b, c)

            >>> a * Product(b, c)
            Product(a, b, c)

            >>> Product(a, b) * Product(c, d)
            Product(a, b, c, d)
        """
        # ensure not to build products of products
        # ensure not to build unions of unions
        lhs_terms = self.args if isinstance(self, Product) else [self]
        rhs_terms = rhs.args if isinstance(rhs, Product) else [rhs]
        return Product(*(lhs_terms + rhs_terms))

    def __pow__(self, n: int) -> "Rule":
        """Exponentiation using the ** syntax.

        Examples:
            >>> A = RuleName("A")
            >>> A ** 3
            Product(A, A, A)

            >>> A ** 1
            A

            >>> A ** 0
            epsilon
        """
        if n < 0:
            raise ValueError("Exponentiation to a negative integer is forbidden.")
        elif n == 0:
            return Epsilon()
        elif n == 1:
            return self
        else:
            return Product(*[self for _ in range(n)])

    def __xor__(self, n: int) -> "Rule":
        """Exponentiation using the ^ syntax.

        Examples:
            >>> A = RuleName("A")
            >>> A ^ 3
            Product(A, A, A)

            >>> A ^ 1
            A

            >>> A ^ 0
            epsilon
        """
        return self**n


class IteratedRule(Rule):
    """The base class for iterated constructions Seq, Set, Cycle, MSet, etc.

    Should not be instantiated directly.
    """

    # This class attribute is used for pretty-printing. It has to be set in
    # classes extending IteratedRule.
    construction_name: str

    def __init__(
        self,
        arg: TypeUnion[Rule, str],
        leq: Optional[int] = None,
        geq: Optional[int] = None,
        eq: Optional[int] = None,
    ):
        """Rule constructor.

        Args:
          arg: a rule describing the elements of the
            collection.

          leq: constrains the collection to have size at most ``leq``.

          geq: constrains the collection to have size at least ``geq``.

          eq: constrains the collection to have size exactly ``eq``.
        """
        super(IteratedRule, self).__init__()

        assert leq is None or geq is None or geq <= leq
        assert eq is None or (leq is None and geq is None)  # eq != 0 => leq == geq == 0

        self.lower_size = None
        self.upper_size = None

        if eq is not None:
            self.lower_size = eq
            self.upper_size = eq
        else:
            self.lower_size = geq
            self.upper_size = leq

        self.arg = _to_rule(arg)

    def __repr__(self) -> str:
        if self.construction_name is None:
            raise ValueError("IteratedRule should not be used directly!")
        constraint = ""
        if self.lower_size is not None:
            if self.lower_size == self.upper_size:
                constraint = ", eq = {}".format(self.upper_size)
                return "{}({}{})".format(self.construction_name, self.arg, constraint)
            else:
                constraint += ", geq = {}".format(self.lower_size)
        if self.upper_size is not None:
            constraint += ", leq = {}".format(self.upper_size)
        return "{}({}{})".format(self.construction_name, self.arg, constraint)

    def markers(self) -> PythonSet["Marker"]:
        """Return the set of all markers in the rule."""
        return self.arg.markers()


class Symbol(Rule):
    """Base class for all grammar symbols (Atom, RuleName, Epsilon).

    Should not be instantiated directly.
    """

    def __init__(self, name: Optional[str] = None):
        super(Symbol, self).__init__()
        if name is None:
            name = _symbol_name_generator.fresh()
        self.name = name

    def __repr__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Symbol) and self.name == other.name

    def __lt__(self, other: Rule) -> bool:
        if not isinstance(other, Symbol):
            raise TypeError("Cannot compare Symbol and {}".format(type(other)))
        return self.name < other.name

    def markers(self) -> PythonSet["Marker"]:
        """Return the set of all markers contained in the expression."""
        return set()


class Marker(Symbol):
    """Marker symbols.

    Examples:
        >>> Marker("u")
        u

        >>> Marker()
        _symbol_0

    Markers are similar to atoms but have size 0. They are usually used to
    mark some special places in the grammar. For instance in the following
    grammar for Motzkin tree, the marker `u` is used to mark unary nodes.

    Examples:
        >>> z, u, M = Atom(), Marker(), RuleName()
        >>> grammar = Grammar({M: z + u * z * M + z * M * M})
    """

    def markers(self) -> PythonSet["Marker"]:
        return {self}


class Singleton:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance


class Epsilon(Singleton, Symbol):
    """The epsilon symbol.

    Epsilon is the class containing only one element of size 0.

    Note:
        There is only one instance of the Epsilon class.

    Examples:
        >>> eps1 = Epsilon()
        >>> eps2 = Epsilon()
        >>> eps1 is eps2
        True
    """

    def __init__(self):
        super(Epsilon, self).__init__("epsilon")


class Atom(Singleton, Symbol):
    """Terminal symbol of a grammar, accounts for the size.

    Atom is the class containing only one element of size 1.

    Note:
        There is only one instance of the Atom class.

    >>> Atom()
    z

    >>> Atom() is Atom()
    True
    """

    def __init__(self):
        super(Atom, self).__init__("z")


class RuleName(Symbol):
    """Non terminal symbols of a grammar.

    Instances of this class represent recursive references to non-terminal
    symbols inside grammar rules. For instance, sequences of atoms ``z`` could
    be defined using the following grammar where the :py:class:`RuleName` ``S``
    refers to itself in its definition:

    Examples:
        >>> epsilon, z, S = Epsilon(), Atom(), RuleName("S")
        >>> Grammar({S: epsilon + z * S})
        {
            S : Union(epsilon, Product(z, S))
        }
    """


def _to_rule(rule: TypeUnion[str, Rule]) -> Rule:
    """Cast strings to RuleNames, otherwise do nothing."""
    if isinstance(rule, Rule):
        return rule
    elif isinstance(rule, str):
        return RuleName(rule)
    else:
        raise ValueError("_to_rule accepts only rules and strings")


def _to_rulename(rule_name: TypeUnion[str, RuleName]) -> RuleName:
    """Cast strings to RuleNames, otherwise do nothing."""
    if isinstance(rule_name, RuleName):
        return rule_name
    elif isinstance(rule_name, str):
        return RuleName(rule_name)
    else:
        raise ValueError("_to_rulename accepts only RuleNames and strings")


class Union(Rule):
    """Disjoint union of two or more rules.

    ``D = Union(A, B, C)`` corresponds to the following grammar in BNF syntax:
    ``D ::= A | B | C``

    Examples:
        >>> A, B, C = RuleName("A"), RuleName("B"), RuleName("C")

        >>> Union(A, B, C)
        Union(A, B, C)

        >>> z = Atom()
        >>> Union(z, A)
        Union(z, A)
    """

    args: List[Rule]

    def __init__(self, *terms: TypeUnion[Rule, str]):
        """Build a union of two or more rules.

        Args:
            terms: list of terms of the union.
        """
        super(Union, self).__init__()

        if len(terms) < 2:
            raise ValueError("Unions should have at least two terms")
        self.args = [_to_rule(arg) for arg in terms]

    def __repr__(self) -> str:
        return "Union({})".format(", ".join(map(repr, self.args)))

    def markers(self) -> PythonSet[Marker]:
        """Return the set of markers contained in the expression."""
        return reduce(lambda x, y: x | y, (arg.markers() for arg in self.args))


class Product(Rule):
    """Cartesian product of two or more rules.

    ``Product(A, B, C)`` corresponds to the following grammar in BNF syntax:
    ``_ ::= A × B × C``

    Examples:
        >>> A, B, C = RuleName("A"), RuleName("B"), RuleName("C")

        >>> Product(A, B, C)
        Product(A, B, C)

        >>> z = Atom()
        >>> Product(z, A)
        Product(z, A)
    """

    args: List[Rule]

    def __init__(self, *factors: TypeUnion[Rule, str]):
        """Build the product of two or more rules.

        Args:
            factors: list of factors of the product.
        """
        super(Product, self).__init__()
        if len(factors) < 2:
            raise ValueError("Products should have at least two components")
        self.args = [_to_rule(arg) for arg in factors]

    def __repr__(self) -> str:
        return "Product({})".format(", ".join(map(repr, self.args)))

    def markers(self) -> PythonSet[Marker]:
        """Return the set of all markers contained in the expression."""
        return reduce(lambda x, y: x | y, (arg.markers() for arg in self.args))


class Seq(IteratedRule):
    """Sequences.

    The Seq construction of the symbolic method models sequences of elements.
    In the following example, Seq is used to represent binary words as
    sequences of bits.

    Example:
        >>> z, one, zero, S = Atom(), Marker("1"), Marker("0"), RuleName("S")
        >>> Grammar({S: Seq(z * (one + zero))})
        {
          S : Seq(Product(z, Union(1, 0)))
        }

    The number of terms of a sequence can be constrained to be greater or
    smaller that some integer constants:

    Examples:
        >>> z, one, zero = Atom(), Marker("1"), Marker("0")

        >>> Seq(z * (one + zero), leq=10, geq=5)
        Seq(Product(z, Union(1, 0)), geq = 5, leq = 10)

        >>> Seq(z * (one + zero), leq=10)
        Seq(Product(z, Union(1, 0)), leq = 10)

        >>> Seq(z * (one + zero), geq=5)
        Seq(Product(z, Union(1, 0)), geq = 5)
    """

    construction_name = "Seq"  # For pretty-printing.


class Set(IteratedRule):
    """Labelled sets.

    The labelled Set construction of the symbolic method models sets of
    elements. In the following example, Set is used to model labelled general
    trees.

    Example:
        >>> z, T = Atom(), RuleName("T")
        >>> Grammar({T: z * Set(T)})
        {
          T : Product(z, Set(T))
        }

    The number of elements of the set can be constrained to be greater or
    smaller that some integer constants:

    Examples:
        >>> z = Atom()

        >>> # sets of 5 to 10 elements
        >>> Set(z, geq=5, leq=10)
        Set(z,  geq = 5, leq = 10)

        >>> # sets of at least 5 elements
        >>> Set(z, geq = 5)
        Set(z, geq = 5)

        >>> # sets of at most 10 elements
        >>> Set(z, leq=10)
        Set(z, leq = 10)
    """

    construction_name = "Set"  # For pretty-printing.


class Cycle(IteratedRule):
    """Labelled cycles.

    A cycle is like a sequence whose components can be cyclically shifted. For
    instance: [a, b, c], [b, c, a] and [c, a, b] represent the same cycle. In
    the following example, Cycle is used to represent the class of permutations
    as a set of cycles:

    Example:
        >>> z, P = Atom(), RuleName("P")
        >>> Grammar({P: Set(Cycle(z))})
        {
          P : Set(Cycle(z))
        }

    The number of elements of the cycle can be constrained to be greater or
    smaller that some integer constants:

    Examples:
        >>> z = Atom()

        >>> Cycle(z, geq=5, leq=10)
        Cycle(z, geq = 5, leq = 10)

        >>> Cycle(z, geq=5)
        Cycle(z, geq = 5)

        >>> Cycle(z, leq=10)
        Cycle(z, leq = 10)
    """

    construction_name = "Cycle"

    def __init__(
        self,
        arg: TypeUnion[Rule, str],
        leq: Optional[int] = None,
        geq: Optional[int] = None,
        eq: Optional[int] = None,
    ):
        if leq == 0 or geq == 0:
            raise ValueError("Cycles should contain at least one component")
        super(Cycle, self).__init__(arg, leq, geq, eq)


class MSet(IteratedRule):
    """Unlabelled multi-sets.

    A multi-set is like a set where elements can occur multiple times. In the
    following example, MSet is used to represent the class of non-plane
    general trees:

    Examples:
        >>> z, T = Atom(), RuleName("T")
        >>> Grammar({T: z * MSet(T)})
        {
          T : Product(z, MSet(T))
        }

    The number of elements of the multi-set can be constrained to be greater or
    smaller that some integer constants:

    Examples:
        >>> z = Atom()

        >>> MSet(z, geq=5, leq=10)
        MSet(z, geq = 5, leq = 10)

        >>> MSet(z, geq=5)
        MSet(z, geq = 5)

        >>> MSet(z, leq=10)
        MSet(z, leq = 10)
    """

    construction_name = "MSet"  # For pretty-printing.


class UCycle(IteratedRule):
    """Unlabelled cycles.

    A cycle is like a sequence whose components can be cyclically shifted. For
    instance: [a, b, c], [b, c, a] and [c, a, b] represent the same cycle.

    The number of elements of the cycle can be constrained to be greater or
    smaller that some integer constants:

    Examples:
        >>> z = Atom()

        >>> UCycle(z, geq=5, leq=10)
        UCycle(z, geq = 5, leq = 10)

        >>> UCycle(z, geq=5)
        UCycle(z, geq = 5)

        >>> UCycle(z, leq=10)
        UCycle(z, leq = 10)
    """

    construction_name = "UCycle"  # For pretty-printing

    def __init__(
        self,
        arg: TypeUnion[Rule, str],
        leq: Optional[int] = None,
        geq: Optional[int] = None,
        eq: Optional[int] = None,
    ):
        if leq == 0 or geq == 0:
            raise ValueError("UCycles should contain at least one component")
        super(UCycle, self).__init__(arg, leq, geq, eq)


class Grammar:
    """Context free grammars."""

    rules: Dict[RuleName, Rule]

    def __init__(
        self,
        rules: Optional[Dict[RuleName, Rule]] = None,
        labelled: bool = False,
    ):
        r"""Create a grammar.

        The rules of the grammar can be either specified at grammar creation by
        feeding them to the constructor or specified later using the
        :py:meth:`add_rule` method.

        Args:
          rules: dictionary mapping RuleNames to Rules

        Examples:
            >>> z = Atom()

            The grammar of sequences

            >>> S = RuleName("S")
            >>> Grammar({S: Epsilon() + z * S})
            {
              S : Union(epsilon, Product(z, S))
            }

            The grammar of binary trees

            >>> B = RuleName("B")
            >>> Grammar({B: Epsilon() + z * B * B})
            {
              B : Union(epsilon, Product(z, B, B))
            }

            The grammar of unary-binary trees (or Motzkin trees)

            >>> g = Grammar()
            >>> T, U, B = RuleName("T"), RuleName("U"), RuleName("B")
            >>> g.add_rule(T, z + U + B)
            >>> g.add_rule(U, z * T)
            >>> g.add_rule(B, z * T * T)
            >>> g
            {
              B : Product(z, T, T),
              T : Union(z, U, B),
              U : Product(z, T)
            }
        """
        self.rules = {}
        rules = rules or {}
        # Pass the dictionary of rules to the add_rule method, add_rule does
        # all the sanitizing work
        for name, rule in rules.items():
            self.add_rule(name, rule)
        self.labelled = labelled

    def add_rule(
        self, rule_name: TypeUnion[RuleName, str], rule: TypeUnion[Rule, str]
    ) -> None:
        """Add a rule to the grammar.

        Args:
            rule_name: a non-terminal symbol. If it was already defined in the grammar,
                it is replaced.

            rule: the rule defining ``rule_name``.

        Examples:
            >>> A, B, C = RuleName("A"), RuleName("B"), RuleName("C")
            >>> g = Grammar()
            >>> g.add_rule(A, Union(B, C))
            >>> g.rules[A]
            Union(B, C)
        """
        self.rules[_to_rulename(rule_name)] = _to_rule(rule)

    def __repr__(self) -> str:
        return (
            "{\n"
            + ",\n".join(
                "  {} : {}".format(non_terminal, expr)
                for non_terminal, expr in sorted(self.rules.items())
            )
            + "\n}"
        )

    def markers(self) -> PythonSet[Marker]:
        """Return all the markers appearing in the grammar.

        Examples:
            >>> z, u = Atom(), Marker("u")
            >>> B = RuleName("B")
            >>> g = Grammar({B: Union(u, Product(z, B, B))})
            >>> g.markers() == {u}
            True
        """
        return reduce(
            lambda x, y: x | y, (expr.markers() for expr in self.rules.values())
        )



# distutils: language = c++
# cython: language_level = 3

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""Boltzmann generator for Context-free grammars.

This module provides functions for generating combinatorial objects (i.e.
objects described by a combinatorial specification, see
:mod:`usainboltz.grammar`) according to the Boltzmann
distribution.

Given an unlabelled combinatorial class A, the Boltzmann distribution of
parameter :math:`x` is such that an object of size n is drawn with the
probability :math:`\frac{x^n}{A(x)}` where :math:`A(x)` denotes the ordinary
generating function of A.  For labelled classes, this probability is set to
:math:`\frac{x^n}{n!A(x)}` where :math:`A(x)` denotes the exponential
generating function of A. See [DuFlLoSc04]_ for details.

By default, the objects produced by the generator are nested tuples of strings
(the atoms). For instance ``('z', ('z', 'epsilon', 'epsilon'), ('z', 'epsilon', 'epsilon'))`` is a
balanced binary tree with 3 internal nodes (z) and 4 leaves (e). To alter this
behaviour and generate other types of objects, you can specify a builder
function for each type of symbol in the grammar. The behaviour of the builders
is that they are applied in bottom up order to the structure "on the fly" during
generation. For instance, in order to generate Dyck words using the grammar for
binary trees, one can use a builder that return ``""`` for each leaf ``"(" +
left child + ")" + right child`` for each node. The builders will receive a
tuple for each product, a string for each atom and builders for unions should be
computed using the :func:`union_builder` helper. See the example below for the
case of Dyck words.

Examples:
    >>> from usainboltz import *
    >>> from usainboltz.generator import rng_seed
    >>> rng_seed(0xDEADBEEFCA7B0172)

    Both binary trees and Dyck words can be generated from the following
    grammar

    >>> epsilon, z, B = Epsilon(), Atom(), RuleName("B")
    >>> grammar = Grammar({B: epsilon + z * B * B})
    >>> generator = Generator(grammar, B, singular=True)

    In order to build Dyck words, one can interpret epsilon as the empty word
    and nodes as the derivation: ``D -> ( D ) D``

    >>> def leaf_builder(_):
    ...     return ""

    >>> def node_builder(tuple):
    ...     _, left, right = tuple
    ...     return "(" + left + ")" + right

    >>> generator.set_builder(B, union_builder(leaf_builder, node_builder))
    >>> res = generator.sample((10, 20))
    >>> dyck_word = res.obj
    >>> dyck_word
    '(())(((((()(((()())))))))())()'
    >>> len(dyck_word) in range(20, 41)
    True

    If on the contrary we want to see trees as S-expressions, we can interpret
    epsilon as the leaf ``"leaf"`` and nodes as ``(node left_child
    right_child)``

    >>> def leaf_builder(_):
    ...     return "leaf"

    >>> def node_builder(tuple):
    ...     _, left, right = tuple
    ...     return "(node {} {})".format(left, right)

    >>> generator.set_builder(B, union_builder(leaf_builder, node_builder))
    >>> res = generator.sample((10, 20))
    >>> print(res.obj)
    (node
      (node
        leaf
        (node
          leaf
          (node
            (node
              (node
                (node leaf leaf)
                leaf)
              (node
                (node
                    leaf
                    (node
                      (node
                        (node leaf leaf)
                        (node
                          (node
                            leaf
                            (node
                              (node leaf leaf)
                              leaf))
                          leaf))
                      leaf))
                leaf))
            (node
              (node leaf leaf)
              leaf))))
      leaf)

    Note that the builders mechanism can also be used to compute some
    statistics on the fly without explicitly building the whole structure. For
    instance the following example illustrates how to generate the **height**
    of a tree following the Boltzmann distribution without building the tree.

    >>> def leaf_height(_):
    ...     return 0

    >>> def node_height(tuple):
    ...     _, left, right = tuple
    ...     return 1 + max(left, right)

    >>> generator.set_builder(B, union_builder(leaf_height, node_height))
    >>> res = generator.sample((10, 20))
    >>> res.obj
    13
    >>> res.sizes[Atom()]
    16

For labelled grammar, the procedure is the same. In this case, the default
builders also output the labelling, they do not only generate the structure.

Example:
    >>> z = Atom()
    >>> e, M = Epsilon(), RuleName()
    >>> g = Grammar({M: e + z * M + z * M * M}, labelled=True)
    >>> generator = Generator(g, M, singular=True)
    >>> res = generator.sample((5,10))  # random
    >>> tree = res.obj
    >>> tree
    (4, (5, (0, (6, (3, (7, (1, 'epsilon', 'epsilon'), 'epsilon'), 'epsilon')), (2, 'epsilon', 'epsilon')), 'epsilon'))

"""

from libc.math cimport exp
from libc.stdint cimport uint64_t
from libcpp.pair cimport pair
from libcpp.stack cimport stack
from libcpp.vector cimport vector

import copy
from collections import namedtuple

from usainboltz.grammar import (
    Atom,
    Epsilon,
    Marker,
    MSet,
    Product,
    RuleName,
    Seq,
    Set,
    Union,
)
from usainboltz.oracle import build_oracle

from usainboltz.cpp_simulator cimport (
    ATOM,
    EPSILON,
    MARKER,
    MSET,
    PRODUCT,
    REF,
    SEQ,
    SET,
    UNION,
    CRule,
    initial_partition,
    next_partition,
    rule_type,
)
from usainboltz.simulator cimport (
    Simulator,
    bounded_geometric,
    bounded_mset_distribution,
    bounded_poisson,
    rng_double,
    rng_get_state,
    rng_i64,
    rng_seed as _rng_seed,
    rng_set_state,
)


# XXX. Is this indirection necessary?
cpdef void rng_seed(uint64_t s):
    _rng_seed(s)


# ------------------------------------------------------- #
# Grammar preprocessing
# ------------------------------------------------------- #

# For performance reasons, we use integers rather that strings to identify
# symbols during generation.
# The Mapping class computes a mapping from integers to rule names, a mapping from
# integers to markers, and their inverse mappings.
#
# All of this should remain hidden from the end user.

class Mapping:
    def __init__(self, grammar):
        # Rule names
        rule_name_to_id, id_to_rule_name = self._make_mapping(grammar.rules.keys())
        self.rule_name_to_id = rule_name_to_id
        self.id_to_rule_name = [id_to_rule_name[i] for i in range(len(id_to_rule_name))]
        # Markers
        marker_to_id, id_to_marker = self._make_mapping(grammar.markers())
        self.marker_to_id = marker_to_id
        self.id_to_marker = [id_to_marker[i] for i in range(len(id_to_marker))]

    def nb_rules(self):
        return len(self.id_to_rule_name)

    def nb_markers(self):
        return len(self.id_to_marker)

    @staticmethod
    def _make_mapping(iterable, start=0):
        x_to_id = {}
        id_to_x = {}
        for i, x in enumerate(iterable, start=start):
            x_to_id[x] = i
            id_to_x[i] = x
        return x_to_id, id_to_x


# ------------------------------------------------------- #
# Generic builders
# ------------------------------------------------------- #

# When it chooses one possible derivation for a union rule, the random
# generator wraps the generated object in a tuple of the form ``(choice_id,
# object)`` so that the builder that will be called on this value has the
# information of which derivation was chosen. This helper function hides this
# machinery to the end user, allowing her to compose builders in a "high-level"
# manner.

def union_builder(*builders):
    """Factory for generating builders for union rules.

    The recommended way to write a builder for a union rule is to use this
    helper: write a auxilliary builder for each component of the union and
    compose them with ``union_builder``.

    Args:
        builders (List[function]): builder functions, one for each component of
            the union.

    Returns:
        function: a builder function for a disjoint union. The resulting
            function applies one of the functions passed as arguments to its
            input depending of the component of the union its argument belongs
            to.

    Examples:
        Assume the symbol ``D`` is defined by ``Union(A, B, C)``, then
        defining a builder for D looks like:

        >>> def build_A(args):
        ...     # Do something
        ...     pass

        >>> def build_B(args):
        ...     # Do something
        ...     pass

        >>> def build_C(args):
        ...     # Do something
        ...     pass

        >>> build_D = union_builder(build_A, build_B, build_C)

        For instance, for binary trees defined by ``B = Union(leaf, Product(z,
        B, B))``, this could be:

        >>> def build_leaf(_):
        ...     return BinaryTree()

        >>> def build_node(args):
        ...     z, left, right = args
        ...     return BinaryTree([left, right])

        >>> build_binarytree = union_builder(build_leaf, build_node)

    """
    def build(obj):
        index, content = obj
        builder = builders[index]
        return builder(content)
    return build

# The following functions generate the default builders for labelled and
# unlabelled structures. These builders produce nested tuples. Choices ids are
# omitted for readability.

cdef inline identity(x):
    return x

cdef inline ProductBuilder(builders):
    def build(terms):
        return tuple([builders[i](terms[i]) for i in range(len(terms))])
    return build

cdef inline SeqBuilder(builder):
    def build(terms):
        return [builder(term) for term in terms]
    return build

cdef make_default_builder(rule):
    """Generate the default builders for a rule.

    Args:
        rule (Rule): a grammar rule.

    Returns:
        function: a simple builder for the rule passed as argument that
            generate objects as nested tuples.
    """
    # Symbols
    if isinstance(rule, RuleName):
        return identity
    elif isinstance(rule, (Atom, Marker, Epsilon)):
        return identity
    # Grammar combinators
    elif isinstance(rule, Product):
        subbuilders = [make_default_builder(component) for component in rule.args]
        return ProductBuilder(subbuilders)
    elif isinstance(rule, Seq):
        subbuilder = make_default_builder(rule.arg)
        return SeqBuilder(subbuilder)
    elif isinstance(rule, Set):
        subbuilder = make_default_builder(rule.arg)
        # Same as Seq
        return SeqBuilder(subbuilder)
    elif isinstance(rule, MSet):
        subbuilder = make_default_builder(rule.arg)
        # Same as Seq
        return SeqBuilder(subbuilder)
    elif isinstance(rule, Union):
        subbuilders = [make_default_builder(component) for component in rule.args]
        return union_builder(*subbuilders)
    else:
        raise NotImplementedError("make_default_builder({})".format(type(rule)))


# ------------------------------------------------------- #
# Random generation
# ------------------------------------------------------- #

# The generator is implemented as a stack machine.
# This stack machine has more instructions that the simulator because
# it also has to build a data structure. We thus split its instructions
# into two kinds:
# - generation instructions describe what to sample (as in the simulator);
# - build instructions describe how to construct the data structure that will be
#   returned to the user.

# Special instructions used by the random generators.
# They all take an integer argument.
ctypedef enum build_instr_name:
    TUPLE,       # Wrap the topmost k elements of the generated stack in a tuple
    LIST,        # Wrap the topmost k elements of the generated stack in a list
    FUNCTION,    # Call the k-th builder on the top of the stack
    WRAP_CHOICE  # Wrap the top element (x) of stack as a pair (k, x)
    MULTIPLICITY # Update the value of the `multiplicity` register of the state machine
    DUPLICATE    # Duplicate the last generated structure several times

# A build instruction and its integer argument
ctypedef struct build_instr:
    build_instr_name name
    int payload

# Union of the two kinds of instructions
ctypedef union instruction_desc:
    build_instr as_build_instr  # Build instruction
    const CRule *as_gen_instr   # Generation instruction (just the rule to generate)

# The instruction type
ctypedef struct instruction:
    bint is_gen_instr
    instruction_desc desc

cdef instruction make_gen_instr(const CRule *rule):
    cdef instruction instruction
    instruction.is_gen_instr = True
    instruction.desc.as_gen_instr = rule
    return instruction

cdef instruction make_build_instr(build_instr_name name, int payload):
    cdef instruction instruction
    instruction.is_gen_instr = False
    instruction.desc.as_build_instr.name = name
    instruction.desc.as_build_instr.payload = payload
    return instruction

# Free Boltzmann sampler (actual generation) of *unlabelled* structures.
# Does not handle the labelling of labelled structures

cdef c_generate(const CRule *first_rule, list builders, list marker_names, list labels):
    cdef list generated = []      # Stack of generated structures
    cdef stack[instruction] todo  # Stack of instructions

    cdef instruction current  # Current instruction
    cdef const CRule* rule    # The next rule to generate (when current.is_gen_instr)
    cdef build_instr instr    # The next build step (when not current.is_gen_instr)

    # A bunch of (many too many?) temporary variables.
    cdef instruction arg_instr
    cdef const CRule* arg
    cdef vector[CRule*] args
    cdef vector[vector[double]] weights
    cdef double r
    cdef unsigned int i = 0, k, multiplicity = 1
    cdef int lower_size, upper_size
    cdef vector[pair[int, int]] partition

    # Pre compute some instructions
    cdef vector[instruction] build_instructions;
    for i in range(len(builders)):
        build_instructions.push_back(make_build_instr(FUNCTION, i))

    # Initialise the stack machine with the first instruction to generate
    todo.push(make_gen_instr(first_rule))

    while not todo.empty():
        current = todo.top()
        todo.pop()
        # Handling of the random generation instructions.
        if current.is_gen_instr:
            rule = current.desc.as_gen_instr
            # Symbols
            if rule.get_type() == EPSILON:
                generated.append("epsilon")
            elif rule.get_type() == ATOM:
                if labels is not None:
                    generated.append(labels.pop())
                else:
                    generated.append("z")
            elif rule.get_type() == MARKER:
                generated.append(marker_names[rule.get_marker_id()])
            elif rule.get_type() == REF:
                todo.push(build_instructions[rule.get_ref_id()])
                todo.push(make_gen_instr(rule.get_ref_rule()))
            # Grammar combinators
            elif rule.get_type() == UNION:
                r = rng_double()
                i = 0
                args = rule.get_union_args()
                weights = rule.get_union_weights()
                while i < args.size():
                    r -= weights[i][multiplicity]
                    if r <= 0:
                        todo.push(make_build_instr(WRAP_CHOICE, i))
                        todo.push(make_gen_instr(args[i]))
                        break
                    i += 1
            elif rule.get_type() == PRODUCT:
                args = rule.get_product_args()
                todo.push(make_build_instr(TUPLE, args.size()))
                for arg in args:
                    todo.push(make_gen_instr(arg))
            elif rule.get_type() == SEQ:
                k = bounded_geometric(
                    1.0 - rule.get_iter_arg_weight()[multiplicity],
                    rule.get_iter_lower_size(),
                    rule.get_iter_upper_size()
                )
                todo.push(make_build_instr(LIST, k))
                arg_instr = make_gen_instr(rule.get_iter_arg())
                for i in range(k):
                    todo.push(arg_instr)
            elif rule.get_type() == SET:
                arg = rule.get_iter_arg()
                k = bounded_poisson(
                    rule.get_iter_weight()[multiplicity],
                    rule.get_iter_arg_weight()[multiplicity],
                    rule.get_iter_lower_size(),
                    rule.get_iter_upper_size()
                )
                # store sets as lists
                todo.push(make_build_instr(LIST, k))
                arg_instr = make_gen_instr(arg)
                for i in range(k):
                    todo.push(arg_instr)
            elif rule.get_type() == MSET:
                lower_size = rule.get_iter_lower_size()
                upper_size = rule.get_iter_upper_size()

                # Save the current multiplicity
                todo.push(make_build_instr(MULTIPLICITY, multiplicity))

                # List of the recursive calls to make (with multiplicity)
                mults = []

                if upper_size >= 0:
                    if lower_size < upper_size:
                        raise NotImplementedError("MSet with bounds other than `eq=`")

                    # Draw a partition under the distribution induced by their weights:
                    partition = initial_partition(lower_size)
                    r = rng_double() * rule.get_iter_weight()[multiplicity];
                    weights = rule.get_mset_partitions_weight()
                    for pw in weights:
                        r -= pw[multiplicity]
                        if r <= 0:
                            break
                        next_partition(partition)

                    # Set up the recursive calls for that partition:
                    for p in partition:
                        # Draw p.second independent elements with multiplicity p.first
                        mults.append((p.first, p.second))

                    # Instruct the generator to wrap the result as a list
                    todo.push(make_build_instr(LIST, lower_size))

                elif rule.get_iter_lower_size() > 0:
                    assert False

                else:
                    # Generate the number or arguments of each multiplicity
                    k = bounded_mset_distribution(
                        rule.get_iter_weight(),
                        rule.get_iter_arg_weight(),
                        multiplicity,
                    )
                    totsize = 0
                    for j in range(1, k):
                        # i ~ Poisson(A(z^{j m}) / j)
                        r = rule.get_iter_arg_weight()[j * multiplicity] / j
                        i = bounded_poisson(exp(r), r, 0, -1)
                        totsize += i * j * multiplicity
                        mults.append((j, i))
                    if k > 0:
                        # i ~ Poisson_{> 0}(A(z^{k m}) / k)
                        r = rule.get_iter_arg_weight()[k * multiplicity] / k
                        i = bounded_poisson(exp(r) - 1.0, r, 1, -1)
                        totsize += i * k
                        mults.append((k, i))

                    # Instruct the generator to wrap the result as a list
                    todo.push(make_build_instr(LIST, totsize))

                # Generate the generation intructions
                arg_instr = make_gen_instr(rule.get_iter_arg())
                for (j, nb) in mults:
                    # Generate i arguments with multiplicity `j * multiplicity`
                    for _ in range(nb):
                        # The generated structure should be duplicated j times
                        if j > 1:
                            todo.push(make_build_instr(DUPLICATE, j))
                        todo.push(arg_instr)
                    # the multiplicity used in the generating functions is j times the
                    # current multiplicity.
                    todo.push(make_build_instr(MULTIPLICITY, j * multiplicity))
            else:
                assert False
        # Handling of the build instructions.
        else:
            instr = current.desc.as_build_instr
            if instr.name == TUPLE:
                t = tuple(generated[:-instr.payload-1:-1])
                del generated[-instr.payload:]
                generated.append(t)
            elif instr.name == LIST:
                if instr.payload == 0:
                    generated.append([])
                else:
                    t = list(generated[-instr.payload:])
                    del generated[-instr.payload:]
                    generated.append(t)
            elif instr.name == FUNCTION:
                func = builders[instr.payload]
                x = generated.pop()
                generated.append(func(x))
            elif instr.name == WRAP_CHOICE:
                choice = generated.pop()
                generated.append((instr.payload, choice))
            elif instr.name == MULTIPLICITY:
                multiplicity = instr.payload
            elif instr.name == DUPLICATE:
                x = generated.pop()
                for _ in range(instr.payload):
                    generated.append(x)
            else:
                assert False

    obj, = generated
    return obj

cdef list _rand_perm(unsigned int n):
    """Draw an uniform random premutation of size ``n`` using the Fisher-Yates algorithm."""
    cdef unsigned int i, j
    cdef list p = list(range(0,n))

    for i in range(n-1,0,-1):
        j = rng_i64(i)
        p[i], p[j] = p[j], p[i]

    return p

cdef c_search_seed(simulator, first_rule, sizes_windows, labelling=False) except +:
    """Search for a tree in a given size window."""
    max_sizes = [size_max for (_, size_max) in sizes_windows]
    cdef uint64_t[4] backup_state

    # I'd need a do while
    rng_get_state(backup_state)

    sizes = simulator.run(first_rule, max_sizes)
    while not _in_sizes_windows(sizes, sizes_windows):
        # save the random generator's state
        rng_get_state(backup_state)
        sizes = simulator.run(first_rule, max_sizes)

    # Generate the labels BEFORE reseting the RNG
    labels = _rand_perm(sizes[0]) if labelling else None

    # Reset the random generator to the state it was just before the simulation
    rng_set_state(backup_state)
    return labels, sizes


# ------------------------------------------------------- #
# High level interface
# ------------------------------------------------------- #

cdef _in_sizes_windows(sizes, sizes_windows):
    for i, (size_min, size_max) in enumerate(sizes_windows):
        size = sizes[i]
        if size < size_min:
            return False
        if size_max > 0 and size > size_max:
            return False
    return True

class GeneratorConfigError(Exception):
    pass

Result = namedtuple("Result", ["obj", "sizes"])

class Generator:
    """High level interface for Boltzmann samplers."""

    def __init__(
        self,
        grammar,
        rule_name=None,
        singular=None,
        expectations=None,
        oracle=None
    ):
        """Make a Generator out of a grammar.

        Args:
            grammar (Grammar): a combinatorial grammar

            rule_name (RuleName): the rule name of the symbol to generate.

            singular (bool): if set, do singular sampling (use the
                singularity of the generating functions as parameter).

            expectations (Dict[Symbol,number]): this is passed to the oracle to
                skew the distribution in order to target some specific values
                for the expected number of the specified symbols. See
                :py:meth:`OracleFromPaganini.tuning` for more details.

            oracle (Oracle): an oracle for computing the values of some
                generating functions. If not supplied, a default generic oracle
                that should work for most use-cases is automatically generated.

        Examples:

        >>> from usainboltz import *

        Some examples using the grammar for binary trees
        >>> z, B = Atom(), RuleName()
        >>> grammar = Grammar({B: Epsilon() + z * B * B})

        Typical use:
        >>> generator = Generator(grammar, B, singular=True)

        Since their is only one symbol in the grammar, the second argument can be
        omitted:
        >>> generator = Generator(grammar, singular=True)

        """

        self.grammar = copy.deepcopy(grammar)
        self.rule_name = self._guess_rule_name(rule_name)
        # Map all symbols in the grammar to an integer identifier.
        # Use arrays rather than dictionaries.
        self.mapping = Mapping(grammar)
        # If `singular` and the expectations are unspecified, do singular sampling
        # by default.
        if singular is None:
            singular = not(bool(expectations))

        # Set up the oracle.
        oracle = oracle or build_oracle(self.grammar)
        self.oracle_values = oracle.tuning(
            rule=Atom() if singular else self.rule_name,
            expectations=expectations,
            singular=singular
        )

        self.sim = Simulator(self.grammar,
                             self.oracle_values,
                             self.mapping)


        # Initialise the default builders.
        self.builders = [
            make_default_builder(self.grammar.rules[self.mapping.id_to_rule_name[id]])
            for id in range(self.mapping.nb_rules())
        ]

    def _guess_rule_name(self, rule_name):
        if rule_name is not None:
            return rule_name
        rule_names = list(self.grammar.rules.keys())
        return rule_names[0]

    def _sanitize_windows(self, sizes_windows):
        if sizes_windows is None:
            sizes_windows = dict()
        elif isinstance(sizes_windows, tuple):
            sizes_windows = {Atom(): sizes_windows}
        return [
            sizes_windows.get(Atom(), (0, -1))
        ] + [
            sizes_windows.get(self.mapping.id_to_marker[i], (0, -1))
            for i in range(self.mapping.nb_markers())
        ]


    def set_builder(self, non_terminal, func):
        """Set the builder for a non-terminal symbol.

        Args:
            non_terminal (RuleName): the name of the non-terminal symbol.

            func (function): the builder.
        """
        symbol_id = self.mapping.rule_name_to_id[non_terminal]
        self.builders[symbol_id] = func

    def get_builder(self, non_terminal):
        """Retrieve the current builder for a non-terminal symbol.

        Args:
            non_terminal (RuleName): the name of the non-terminal symbol.

        Returns:
            function: the current builder currently bound to the rule name
                passed as argument.
        """
        symbol_id = self.mapping.rule_name_to_id[non_terminal]
        return self.builders[symbol_id]

    def _search_seed(self, sizes_windows):
        return c_search_seed(
            self.sim,
            self.rule_name,
            self._sanitize_windows(sizes_windows),
            self.grammar.labelled,
        )

    def _generate(self, labels):
        # This local variable is necessary for cython to see that self.sim is of type
        # Simulator.
        cdef Simulator sim = self.sim
        cdef CRule* start = sim.c_rules[self.mapping.rule_name_to_id[self.rule_name]]
        return c_generate(start, self.builders, self.mapping.id_to_marker, labels)

    def sample(self, sizes_windows):
        """Generate a term of the grammar given the prescribed sizes.

        Args:
            sizes_windows (Tuple[int,int]] | Dict[(Atom|Marker),Tuple[int,int]]):
                for atoms (resp. markers) for which a size window is
                specified, the generator will only generate objects whose
                number of those atoms (resp. markers) are in the
                specified size windows. For instance a
                ``generator.sample({z: (100, 200)})`` will only generate
                objects with a number of ``z`` belonging to the
                interval `[100, 200]`.
        """
        labels, sizes = self._search_seed(sizes_windows)
        res = self._generate(labels)
        ret_sizes = {
            self.mapping.id_to_marker[i]: sizes[i+1] for i in range(len(sizes)-1)
        }
        ret_sizes[Atom()] = sizes[0]
        return Result(obj=res, sizes=ret_sizes)



# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

"""A Boltzmann sampler for Pólya n-ary trees.

>>> from usainboltz import *
>>> from usainboltz.generator import rng_seed
>>> rng_seed(0xDEADBEEFB017)  # For reproducibility

Pólya trees are unlabelled unordered rooted trees. They are specified using the
:py:class:`~usainboltz.grammar.MSet` operator:

>>> z = Atom()
>>> T = RuleName("T")
>>> grammar = Grammar({T: z * MSet(T)})
>>> grammar
{
  T : Product(z, MSet(T))
}

The singularity of theses trees is near z = 0.33832185689920769 (cf. [FS2009] section
VII.5 page 475) and the value of the generating function at this point is 1.
Let us check that the oracle finds these values:

>>> oracle = build_oracle(grammar)
>>> values = oracle.tuning(z, singular=True)
>>> # NB: values[z][j] stores the value of z^j
>>> print(f"{values[z][1]:.6f}")
0.338322
>>> print(f"{values[T][1]:.6f}")
1.000000

Now, let us generate a Pólya tree:
>>> generator = Generator(grammar)
>>> res = generator.sample((10, 15))
>>> print(res.obj)
('z', [('z', []),
       ('z', [('z', []),
              ('z', []),
              ('z', [('z', []),
                     ('z', [('z', []),
                            ('z', []),
                            ('z', []),
                            ('z', []),
                            ('z', [])])])])])

Finally, let us check that the number of atoms in the structure corresponds to the
size reported by the generator.
>>> def compute_size(tree):
...     _, mset = tree
...     return 1 + sum((compute_size(tree) for tree in mset))
>>> res.sizes[z] == compute_size(res.obj), res.sizes[z]
(True, 13)
"""

from usainboltz import Atom, Generator, Grammar, MSet, RuleName

if __name__ == "__main__":
    # Construct the grammar
    z = Atom()
    T = RuleName("T")
    grammar = Grammar({T: z * MSet(T)})

    # Instatiate the generator
    gen = Generator(grammar)
    for _ in range(10):
        print(gen.sample((20, 30)))



# cython: language_level = 3

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

from libc.stdint cimport int64_t, uint64_t
from libcpp.pair cimport pair
from libcpp.vector cimport vector


cdef extern from "cpp_simulator.cpp":
    pass

cdef extern from "cpp_xoshiro.cpp":
    pass

cdef extern from "cpp_simulator.hpp":
    ctypedef enum rule_type:
        REF,
        ATOM,
        MARKER,
        EPSILON,
        UNION,
        PRODUCT,
        SEQ,
        SET,
        MSET,

    cdef cppclass CRule:
        rule_type get_type()

        int get_ref_id()
        void set_ref_rule(CRule*)
        CRule *get_ref_rule()

        int get_marker_id()

        const vector[CRule*] &get_product_args()

        const vector[CRule*] &get_union_args()
        const vector[vector[double]] &get_union_weights()

        const CRule* get_iter_arg()
        const vector[double] &get_iter_weight()
        const vector[double] &get_iter_arg_weight()
        int get_iter_lower_size()
        int get_iter_upper_size()
        const vector[vector[double]] &get_mset_partitions_weight()

    CRule* make_ref(int)
    CRule* make_atom()
    CRule* make_epsilon()
    CRule* make_marker(int)
    CRule* make_union(vector[CRule*] &args, vector[vector[double]] & weights)
    CRule* make_product(vector[CRule*]& args)
    CRule* make_seq(CRule*, vector[double], vector[double], int, int)
    CRule* make_set(CRule*, vector[double], vector[double], int, int);
    CRule* make_mset(CRule*, vector[double], vector[double], vector[vector[double]],
                     int, int);

    int bnd_geometric(double, int, int)
    int bnd_poisson(double, double, int, int)
    int bnd_mset_dist(vector[double], vector[double], int) except +

    vector[pair[int, int]] initial_partition(int)
    bint next_partition(vector[pair[int, int]]&) except +

    vector[int] c_simulate(CRule*, vector[int]& max_sizes) except +

cdef extern from "cpp_xoshiro.hpp":
    double rand_double()
    int64_t rand_i64(const int64_t bound)

    void seed(const uint64_t seed)
    void get_state(uint64_t dest[4])
    void set_state(const uint64_t s[4])



# coding: utf-8

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

import warnings

from usainboltz.generator import Generator, union_builder
from usainboltz.grammar import (
    Atom,
    Cycle,
    Epsilon,
    Grammar,
    Marker,
    MSet,
    Product,
    RuleName,
    Seq,
    Set,
    UCycle,
    Union,
)
from usainboltz.oracle import OracleFromDict, OracleFromPaganini, build_oracle

__all__ = [
    "Generator",
    "union_builder",
    "Marker",
    "Epsilon",
    "Atom",
    "Cycle",
    "Grammar",
    "MSet",
    "Product",
    "RuleName",
    "Seq",
    "Set",
    "Union",
    "UCycle",
    "build_oracle",
    "OracleFromPaganini",
    "OracleFromDict",
]


# XXX: remove me
# XXX: https://github.com/maciej-bendkowski/paganini/issues/6
warnings.filterwarnings("ignore", category=UserWarning, module="cvxpy.problems.problem")


try:
    from usainboltz.oracle import OracleFromNewtonGF  # noqa: F401

    __all__.append("OracleFromNewtonGF")
except ImportError:
    pass



# coding: utf8

from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup

with open("README.md", "r") as file:
    long_description = file.read()

extra_compile_args = ["-std=c++11", "-Wpedantic", "-Wextra"]

extensions = [
    Extension(
        "usainboltz.cpp_simulator",
        sources=["src/usainboltz/cpp_simulator.cpp", "src/xoshiro/cpp_xoshiro.cpp"],
        include_dirs=["src/xoshiro", "src/usainboltz"],
        language="c++",
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "usainboltz.simulator",
        sources=["src/usainboltz/simulator.pyx"],
        include_dirs=["src/xoshiro", "src/usainboltz"],
        language="c++",
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "usainboltz.generator",
        sources=["src/usainboltz/generator.pyx"],
        include_dirs=["src/xoshiro/", "src/usainboltz"],
        language="c++",
        extra_compile_args=extra_compile_args,
    ),
]

setup(
    name="usainboltz",
    version="0.2.1",
    author="Matthieu Dien, Martin Pépin",
    author_email="kerl@wkerl.me",
    description="Fast Boltzmann random generators for SageMath",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/ParComb/usain-boltz",
    project_urls={
        "Bug Tracker": "https://gitlab.com/ParComb/usain-boltz/issues",
        "Documentation": "https://usain-boltz.readthedocs.io",
        "Source Code": "https://gitlab.com/ParComb/usain-boltz",
    },
    packages=find_packages("src"),
    package_dir={"": "src"},
    ext_modules=cythonize(extensions, language_level=3),
    install_requires=["paganini >= 1.3.3", "cvxpy >= 1.2.0"],
    license="GPLv3+",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Cython",
        "Programming Language :: C++",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



# coding: utf-8

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""
Cayley trees
============

Cayley trees are non-ordered labeled trees. We can use SageMath
`LabelledRootedTree <http://doc.sagemath.org/html/en/reference/combinat/sage/combinat/rooted_tree.html>`_ # noqa
to build them.
We describe here how to write a grammar for Cayley trees in UsainBoltz
and how to generated sage objects.

The grammar of Cayley trees of size given by their number of vertices,
can be written as follows:

>>> from usainboltz import *
>>> from usainboltz.generator import rng_seed
>>> import warnings; warnings.filterwarnings("ignore", category=UserWarning, module='cvxpy.problems.problem')
>>> rng_seed(0xDEADBEEF)  # For reproducibility

>>> z = Atom()
>>> C = RuleName("C")
>>> grammar = Grammar({C: z * Set(C)}, labelled=True)
>>> grammar
{
  C : Product(z, Set(C))
}

To obtain a generator for this grammar, one must write:

>>> generator = Generator(grammar, C, singular=True)

But by default, UsainBoltz generates tuples:

>>> res = generator.sample((10, 20))
>>> res.obj
(1, [(8, [(0, [(7, [(5, [])])])]), (4, [(3, [(6, [])]), (9, [(2, [])])])])

Even if the default data structure generated use list it should be interpreted
as unordered.

Sage LabelledRootedTree can be obtained using the builders feature: we must tell the
generator how to build `C` objects:

>>> from sage.combinat.rooted_tree import LabelledRootedTree

A Cayley tree is a labeled root whose children are Cayley trees

>>> def build_tree(tupl):
...     z, sub_trees = tupl
...     return LabelledRootedTree(sub_trees, label=z)
>>> generator.set_builder(C, build_tree)

Now, the generator directly generates LabelledRootedTree objects:

>>> res = generator.sample((10, 20))
>>> tree = res.obj
>>> tree.parent()
Labelled rooted trees

>>> from sage.all import ascii_art
>>> ascii_art(tree)
      7
      |
      0
      |
      6
      |
      4
      |
      10
      |
      2_
     / /
    3 9
    | |
    8 5
      |
      1
"""



# coding: utf-8

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""Dyck words.

Dyck words is the language over the alphabet ``{'(', ')'}`` described by the
following grammar::

    D ::= <empty word>
        | '(' D ')' D

It is the language of "well-parenthesised" words. In sage, Dyck words are
implemented in the
`dyck_word <https://doc.sagemath.org/html/en/reference/combinat/sage/combinat/dyck_word.html>`_ module. # noqa
The bijection between binary trees and Dyck words allows us to generate directly
Dyck words from the binary tree generator from the
:py:mod:`usainboltz.sage_examples.binary_tree` module using an ad-hoc builder:

>>> from usainboltz import *
>>> from usainboltz.generator import rng_seed
>>> rng_seed(0xDEADBEEF)  # For reproducibility
>>> import warnings; warnings.filterwarnings("ignore", category=UserWarning, module='cvxpy.problems.problem')

>>> # The binary tree sampler
>>> z, leave = Atom(), Epsilon()
>>> B = RuleName()
>>> grammar = Grammar(rules={B: leave + z * B * B})
>>> generator = Generator(grammar, B, singular=True)

>>> # Ad-hoc builder
>>> def empty_word_builder(_):
...     return ""
>>> def product_builder(t):
...     _, left, right = t
...     return "(" + left + ")" + right
>>> dyck_word_builder = union_builder(empty_word_builder, product_builder)
>>> generator.set_builder(B, dyck_word_builder)

>>> res = generator.sample((10, 20))
>>> word = res.obj
>>> word
'()((((((()))))(()))())'

>>> from sage.all import ascii_art, DyckWord
>>> ascii_art(DyckWord(word))
            /\            
           /  \           
          /    \          
         /      \  /\     
        /        \/  \    
       /              \/\ 
    /\/                  \
"""



# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""A Boltzmann sampler for Set partitions (Bell numbers).

>>> from usainboltz.examples.expnums import *
>>> from usainboltz.generator import rng_seed
>>> rng_seed(0xDEADBEEF)
>>> res = generator.sample((10, 20))
>>> res.obj
[[9], [2, 0], [11, 6, 1], [5], [12], [4], [13, 10], [3], [7, 8]]
"""

from math import exp

from usainboltz import Atom, Generator, Grammar, RuleName, Set, build_oracle

z = Atom()
B = RuleName()

grammar = Grammar({B: Set(Set(z, geq=1))}, labelled=True)

# XXX. Paganini does not support Set(z, geq=1) at the moment.
# Expected size is z * exp(z)
# Choose z = 2
oracle = build_oracle({z: 2, B: exp(exp(2) - 1)})
generator = Generator(grammar, B, oracle=oracle)


def build(partition):
    return partition


generator.set_builder(B, build)



# coding: utf-8

# distutils: sources = cpp_xoshiro.cpp
# cython: language_level = 3

# Copyright 2019-2022 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

from libc.stdint cimport int64_t, uint64_t


cdef extern from "cpp_xoshiro.hpp":
    double rand_double()
    int64_t rand_i64(const int64_t bound)

    void seed(const uint64_t seed)
    void get_state(uint64_t dest[4])
    void set_state(const uint64_t s[4])
